import os
import logging
from typing import Callable
import asyncio
import threading
import multiprocessing

from warp_beacon.jobs.types import JobType
from warp_beacon.jobs.upload_job import UploadJob
from warp_beacon.storage import Storage

class AsyncUploader(object):
	__JOE_BIDEN_WAKEUP = None
	threads = []
	allow_loop = True
	job_queue = None
	callbacks = {}
	storage = None
	in_process = set()
	loop = None
	admin_message_callback = None
	request_yt_auth_callback = None
	pool_size = 1

	def __init__(self,
			loop: asyncio.AbstractEventLoop,
			storage: Storage,
			admin_message_callback: Callable,
			request_yt_auth_callback: Callable,
			pool_size: int=min(32, os.cpu_count() + 4)
		) -> None:
		self.storage = storage
		self.loop = loop
		self.job_queue = multiprocessing.Queue()
		self.admin_message_callback = admin_message_callback
		self.request_yt_auth_callback = request_yt_auth_callback
		self.pool_size = pool_size
	
	def __del__(self) -> None:
		self.stop_all()

	def start(self) -> None:
		for _ in range(self.pool_size):
			thread = threading.Thread(target=self.do_work)
			thread.start()
			self.threads.append(thread)

	def add_callback(self, message_id: int, callback: Callable) -> None:
		async def callback_wrap(*args, **kwargs) -> None:
			await callback(*args, **kwargs)
			#self.remove_callback(message_id)
		self.callbacks[message_id] = {"callback": callback_wrap}

	def remove_callback(self, message_id: int) -> None:
		if message_id in self.callbacks:
			logging.debug("Removing callback with message id #%d", message_id)
			del self.callbacks[message_id]

	def stop_all(self) -> None:
		self.allow_loop = False
		if self.threads:
			for i in self.threads:
				t_id = i.native_id
				logging.info("Stopping thread #'%s'", t_id)
				self.job_queue.put(self.__JOE_BIDEN_WAKEUP)
				i.join()
				logging.info("Thread #'%s' stopped", t_id)
		self.threads.clear()

	def is_inprocess(self, uniq_id: str) -> bool:
		return uniq_id in self.in_process
	
	def process_done(self, uniq_id: str) -> None:
		self.in_process.discard(uniq_id)

	def set_inprocess(self, uniq_id: str) -> None:
		self.in_process.add(uniq_id)

	def queue_task(self, job: UploadJob) -> None:
		self.job_queue.put_nowait(job)

	def do_work(self) -> None:
		logging.info("Upload worker started")
		while self.allow_loop:
			try:
				try:
					job = self.job_queue.get()
					if job is self.__JOE_BIDEN_WAKEUP:
						break
					if job.is_message_to_admin and job.message_text and self.admin_message_callback:
						self.loop.call_soon_threadsafe(
							asyncio.create_task,
							self.admin_message_callback(job.message_text, job.account_admins)
						)
						continue
					if job.yt_auth and self.request_yt_auth_callback:
						self.loop.call_soon_threadsafe(
							asyncio.create_task,
							self.request_yt_auth_callback()
						)
						continue

					path = ""
					if job.media_type == JobType.COLLECTION:
						for i in job.media_collection:
							for j in i:
								path += f"{j.local_media_path}; "
					else:
						path = job.local_media_path
	
					in_process = job.in_process
					uniq_id = job.uniq_id
					message_id = job.placeholder_message_id

					if not in_process and not job.job_failed and not job.job_warning and not job.replay:
						logging.info("Accepted upload job, file(s): '%s'", path)

					try:
						if message_id in self.callbacks:
							if job.job_failed:
								logging.info("URL '%s' download failed. Skipping upload job ...", job.url)
								if job.job_failed_msg: # we want to say something to user
									self.loop.call_soon_threadsafe(
										asyncio.create_task,
										self.callbacks[message_id]["callback"](job)
									)
								self.process_done(uniq_id)
								self.remove_callback(message_id)
								continue
							
							if job.replay:
								self.loop.call_soon_threadsafe(
									asyncio.create_task,
									self.callbacks[message_id]["callback"](job)
								)
								self.remove_callback(message_id)
								continue

							if job.job_warning:
								logging.info("Job warning occurred ...")
								if job.job_warning_msg:
									self.loop.call_soon_threadsafe(
										asyncio.create_task,
										self.callbacks[message_id]["callback"](job)
									)
								continue
							if in_process:
								db_list_dicts = self.storage.db_lookup_id(uniq_id)
								if db_list_dicts:
									tg_file_ids = [i["tg_file_id"] for i in db_list_dicts]
									dlds_len = len(db_list_dicts)
									if dlds_len > 1:
										job.tg_file_id = ",".join(tg_file_ids)
										job.media_type = JobType.COLLECTION
									elif dlds_len:
										job.tg_file_id = ",".join(tg_file_ids)
										db_data = db_list_dicts.pop()
										job.media_type = JobType[db_data["media_type"].upper()]
										job.canonical_name = db_data.get("canonical_name", "")
									self.loop.call_soon_threadsafe(
										asyncio.create_task,
										self.callbacks[message_id]["callback"](job)
									)
									self.process_done(uniq_id)
									self.remove_callback(message_id)
								else:
									self.queue_task(job)
							else:
								self.loop.call_soon_threadsafe(
									asyncio.create_task,
									self.callbacks[message_id]["callback"](job)
								)
								self.process_done(uniq_id)
								self.remove_callback(message_id)
						else:
							logging.info("No callback no call!!")
					except Exception as e:
						logging.exception(e)
				except multiprocessing.Queue.empty:
					pass
			except Exception as e:
				logging.error("Exception occurred inside upload worker!")
				logging.exception(e)
		logging.info("Thread done")



from typing import Union, Optional

import re

from pyrogram import Client
from pyrogram.types import Message, ChatMember
from pyrogram import enums

from warp_beacon.jobs import Origin
from warp_beacon.jobs.types import JobType

import logging

class Utils(object):
	expected_patronum_compiled_re = re.compile(r'Expected ([A-Z]+), got ([A-Z]+) file id instead')
	mention_re = re.compile(r'@[A-Za-z0-9_.]{4,}')

	@staticmethod
	def extract_file_id(message: Message) -> Optional[str]:
		possible_attrs = ("video", "photo", "audio", "animation", "document")
		for attr in possible_attrs:
			if hasattr(message, attr):
				_attr = getattr(message, attr, None)
				if _attr:
					tg_id = getattr(_attr, "file_id", None)
					if tg_id:
						return tg_id
		return None

	@staticmethod
	def extract_origin(url: str) -> Origin:
		if "instagram.com/" in url:
			return Origin.INSTAGRAM

		if "youtube.com/" in url and "shorts/" in url:
			return Origin.YT_SHORTS

		if "youtube.com/" in url and "music." in url:
			return Origin.YT_MUSIC

		if "youtu.be/" in url:
			return Origin.YOUTU_BE

		if "youtube.com/" in url:
			return Origin.YOUTUBE

		return Origin.UNKNOWN

	@staticmethod
	def parse_expected_patronum_error(err_text: str) -> tuple:
		'''
			Input example: 'Expected VIDEO, got ANIMATION file id instead'
		'''
		capture = re.match(Utils.expected_patronum_compiled_re, err_text)
		expected_value, got_value = capture[1], capture[2]

		return JobType[expected_value], JobType[got_value]

	@staticmethod
	def chunker(seq: list, size: int) -> list:
		return (seq[pos:pos + size] for pos in range(0, len(seq), size))

	@staticmethod
	def extract_message_text(message: Message) -> str:
		if hasattr(message, "text") and message.text:
			return message.text
		if hasattr(message, "caption") and message.caption:
			return message.caption

		return ''

	@staticmethod
	def extract_message_author(message: Message) -> str:
		#if message.from_user:
		#	if message.from_user.username:
		#		return message.from_user.username
		#	if message.from_user.first_name:
		#		return message.from_user.first_name
		#	if message.from_user.id:
		return str(message.from_user.mention(style=enums.ParseMode.HTML))
		#if message.sender_chat:
		#	if message.sender_chat.username:
		#		return message.sender_chat.username
		#	if message.sender_chat.title:
		#		return message.sender_chat.title
		#return ''

	@staticmethod
	def compute_leftover(urls: list, message: str) -> str:
		msg_leftover = ""
		if len(message) > sum(len(u) for u in urls):
			msg_leftover = message
			for u in urls:
				msg_leftover = msg_leftover.replace(u, '')
		return msg_leftover.strip()
	
	@staticmethod
	async def find_chat_user_by_id(client: Client, chat_id: int, user_id: int) -> Optional[ChatMember]:
		ret = None
		async for member in client.get_chat_members(chat_id):
			user_id = 0
			if member.user.id == user_id:
				ret = member.user
				break
		return ret

	@staticmethod
	async def handle_mentions(chat_id: int, client: Client, message: str) -> str:
		try:
			mentions = Utils.mention_re.findall(message)
			if not mentions:
				return message

			logging.info("Processing chat_id: '%s'", str(chat_id))
			chat = await client.get_chat(chat_id)
			chat_id = chat.id

			chat_member = await client.get_chat_member(chat_id, client.me.id)
			if chat_member.status not in (enums.ChatMemberStatus.ADMINISTRATOR, enums.ChatMemberStatus.OWNER):
				logging.warning("Bot is not an admin in this channel!")
				return message

			username = ''
			members = [member async for member in client.get_chat_members(chat_id)]
			for mention in mentions:
				username = mention[1:].strip()
				if username:
					user_id = None
					display_name = username

					for member in members:
						if member.user.username and member.user.username.lower() == username.lower():
							user_id = member.user.id
							display_name = member.user.first_name or member.user.last_name or username
							break

					if user_id is None:
						for member in members:
							if member.user.first_name and member.user.first_name.lower() == username.lower():
								user_id = member.user.id
								display_name = member.user.first_name
								break

					if user_id:
						message = message.replace(f"@{username}", f'<a href="tg://user?id={user_id}">{display_name}</a>')
		except Exception as e:
			logging.warning("Exception occurred while handling TG mentions!")
			logging.exception(e)
		return message



import logging

from warp_beacon.jobs.types import JobType
from warp_beacon.scraper.youtube.abstract import YoutubeAbstract
from warp_beacon.scraper.exceptions import NotFound

from warp_beacon.mediainfo.video import VideoInfo

class YoutubeShortsScraper(YoutubeAbstract):
	YT_MAX_RETRIES_DEFAULT = 8
	YT_PAUSE_BEFORE_RETRY_DEFAULT = 3
	YT_TIMEOUT_DEFAULT = 2
	YT_TIMEOUT_INCREMENT_DEFAULT = 60

	def _download(self, url: str, session: bool = True, timeout: int = 0) -> list:
		res = []
		thumbnail = None
		yt = self.build_yt(url, session=session)
		stream = yt.streams.get_highest_resolution()

		if not stream:
			raise NotFound("No suitable video stream found")

		local_file = stream.download(
			output_path=self.DOWNLOAD_DIR,
			max_retries=0,
			timeout=timeout,
			skip_existing=False,
			filename_prefix="yt_download_"
		)

		local_file = self.rename_local_file(local_file)
		vinfo = VideoInfo(local_file)
		thumbnail = self.download_hndlr(self.download_thumbnail, video_id=yt.video_id, crop_center=vinfo.get_demensions())

		logging.debug("Temp filename: '%s'", local_file)
		res.append({
			"local_media_path": local_file,
			"performer": yt.author,
			"thumb": thumbnail,
			"canonical_name": stream.title,
			"media_type": JobType.VIDEO
		})

		return res

	def _download_yt_dlp(self, url: str, timeout: int = 60) -> list:
		res = []
		thumbnail = None
		video_id = self.get_video_id(url)
		if video_id:
			thumbnail = self.download_hndlr(self.download_thumbnail, video_id)
		with self.build_yt_dlp(timeout) as ydl:
			info = ydl.extract_info(url, download=True)
			local_file = ydl.prepare_filename(info)
			logging.debug("Temp filename: '%s'", local_file)
			res.append({
				"local_media_path": local_file,
				"performer": info.get("uploader", "Unknown"),
				"thumb": thumbnail,
				"canonical_name": info.get("title", ''),
				"media_type": JobType.VIDEO
			})

		return res


import time
import os
import logging

import av
from pytubefix.exceptions import AgeRestrictedError

from warp_beacon.jobs.types import JobType
from warp_beacon.scraper.youtube.abstract import YoutubeAbstract
from warp_beacon.scraper.exceptions import YoutubeLiveError, NotFound, YotubeAgeRestrictedError

class YoutubeScraper(YoutubeAbstract):
	YT_MAX_RETRIES_DEFAULT = 8
	YT_PAUSE_BEFORE_RETRY_DEFAULT = 3
	YT_TIMEOUT_DEFAULT = 2
	YT_TIMEOUT_INCREMENT_DEFAULT = 60

	def is_live(self, data: dict) -> bool:
		'''
		x.contents.twoColumnWatchNextResults.results.results.contents[0].videoPrimaryInfoRenderer.viewCount.videoViewCountRenderer.isLive
		'''
		try:
			contents = data.get("contents", {}).get("twoColumnWatchNextResults", {}).get("results", {}).get("results", {}).get("contents", [])
			for i in contents:
				video_view_count_renderer = i.get("videoPrimaryInfoRenderer", {}).get("viewCount", {}).get("videoViewCountRenderer", {})
				if video_view_count_renderer:
					return video_view_count_renderer.get("isLive", False)
		except Exception as e:
			logging.warning("Failed to check if stream is live!")
			logging.exception(e)

		return False

	def _download(self, url: str, session: bool = True, timeout: int = 60) -> list:
		res = self._download_pytubefix_max_res(url=url, session=session, timeout=timeout)
		if not res:
			res = self._download_pytube_dash(url=url, session=session, timeout=timeout)

		return res
	
	def mux_raw_copy(self, video_path: str, audio_path: str, output_path: str) -> str:
		try:
			with av.open(video_path) as input_video, av.open(audio_path) as input_audio, av.open(output_path, mode='w') as output:
				in_video_stream = input_video.streams.video[0]
				in_audio_stream = input_audio.streams.audio[0]

				video_stream_map = output.add_stream(template=in_video_stream)
				audio_stream_map = output.add_stream(template=in_audio_stream)

				for packet in input_video.demux(in_video_stream):
					if packet.dts is None:
						continue
					packet.stream = video_stream_map
					output.mux(packet)

				for packet in input_audio.demux(in_audio_stream):
					if packet.dts is None:
						continue
					packet.stream = audio_stream_map
					output.mux(packet)
		except Exception as e:
			logging.error("Failed to mux audio and video!")
			logging.exception(e)
			return ''

		return output_path

	def _download_pytubefix_max_res(self, url: str, session: bool = True, timeout: int = 60) -> list:
		res = []
		local_video_file, local_audio_file = '', ''
		try:
			thumbnail = None
			video_id = self.get_video_id(url)
			if video_id:
				thumbnail = self.download_hndlr(self.download_thumbnail, video_id)
			yt = self.build_yt(url, session=session)

			if self.is_live(yt.initial_data):
				raise YoutubeLiveError("Youtube Live is not supported")
			
			video_stream = yt.streams.filter(adaptive=True, file_extension='mp4', only_video=True).order_by('resolution').desc().first()
			audio_stream = yt.streams.filter(adaptive=True, file_extension='mp4', only_audio=True).order_by('abr').desc().first()

			local_video_file = video_stream.download(
				output_path=self.DOWNLOAD_DIR,
				max_retries=3,
				timeout=timeout,
				skip_existing=False,
				filename_prefix="yt_download_video_"
			)
			local_video_file = self.rename_local_file(local_video_file)
			logging.debug("Temp video filename: '%s'", local_video_file)
			local_audio_file = audio_stream.download(
				output_path=self.DOWNLOAD_DIR,
				max_retries=3,
				timeout=timeout,
				skip_existing=False,
				filename_prefix="yt_download_audio_"
			)
			local_audio_file = self.rename_local_file(local_audio_file)
			logging.debug("Temp audio filename: '%s'", local_audio_file)

			muxed_video = self.mux_raw_copy(
				video_path=local_video_file,
				audio_path=local_audio_file,
				output_path=f"{self.DOWNLOAD_DIR}/yt_muxed_video_{int(time.time())}.mp4")
			if muxed_video:
				muxed_video = self.rename_local_file(muxed_video)
				logging.debug("Temp muxed filename: '%s'", muxed_video)

				res.append({
					"local_media_path": muxed_video,
					"performer": yt.author,
					"thumb": thumbnail,
					"canonical_name": yt.title,
					"media_type": JobType.VIDEO
				})
		except AgeRestrictedError as e:
			raise YotubeAgeRestrictedError("Youtube Age Restricted error")
		finally:
			if os.path.exists(local_video_file):
				os.unlink(local_video_file)
			if os.path.exists(local_audio_file):
				os.unlink(local_audio_file)
		
		return res

	def _download_pytube_dash(self, url: str, session: bool = True, timeout: int = 60) -> list:
		res = []
		try:
			thumbnail = None
			video_id = self.get_video_id(url)
			if video_id:
				thumbnail = self.download_hndlr(self.download_thumbnail, video_id)
			yt = self.build_yt(url, session=session)

			if self.is_live(yt.initial_data):
				raise YoutubeLiveError("Youtube Live is not supported")

			stream = yt.streams.get_highest_resolution()

			if not stream:
				raise NotFound("No suitable video stream found")

			logging.info("Starting download ...")

			local_file = stream.download(
				output_path=self.DOWNLOAD_DIR,
				max_retries=0,
				timeout=timeout,
				skip_existing=False,
				filename_prefix="yt_download_"
			)
			logging.debug("Temp filename: '%s'", local_file)
			res.append({
				"local_media_path": self.rename_local_file(local_file),
				"performer": yt.author,
				"thumb": thumbnail,
				"canonical_name": stream.title,
				"media_type": JobType.VIDEO
			})
		except AgeRestrictedError as e:
			raise YotubeAgeRestrictedError("Youtube Age Restricted error")

		return res
	
	def _download_yt_dlp(self, url: str, timeout: int = 60) -> list:
		res = []
		thumbnail = None
		video_id = self.get_video_id(url)
		if video_id:
			thumbnail = self.download_hndlr(self.download_thumbnail, video_id)
		with self.build_yt_dlp(timeout) as ydl:
			info = ydl.extract_info(url, download=True)
			local_file = ydl.prepare_filename(info)
			logging.debug("Temp filename: '%s'", local_file)
			res.append({
				"local_media_path": local_file,
				"performer": info.get("uploader", "Unknown"),
				"thumb": thumbnail,
				"canonical_name": info.get("title", ''),
				"media_type": JobType.VIDEO
			})

		return res